# -*- coding: utf-8 -*-
"""
Implementation of UNet with Inception Convolutions - UInc
"""

import torch.nn.functional as F
import torch
from GANDLF.models.seg_modules.ResNetModule import ResNetModule
from GANDLF.models.seg_modules.InceptionModule import InceptionModule
from GANDLF.models.seg_modules.IncDownsamplingModule import IncDownsamplingModule
from GANDLF.models.seg_modules.IncUpsamplingModule import IncUpsamplingModule
from GANDLF.models.seg_modules.IncConv import IncConv
from GANDLF.models.seg_modules.IncDropout import IncDropout
from .modelBase import ModelBase
from GANDLF.utils.generic import checkPatchDivisibility


class uinc(ModelBase):
    """
    This is the implementation of the following paper: https://arxiv.org/abs/1907.02110.
    Please look at the seg_module files (towards the end), to get  a better sense of
    the Inception Module implemented. The res parameter is for the addition of the initial feature
    map with the final feature map after performance of the convolution. For the decoding module,
    not the initial input but the input after the first convolution is added to the final output
    since the initial input and the final one do not have the same dimensions.
    """

    def __init__(
        self,
        parameters: dict,
    ):
        super(uinc, self).__init__(parameters)

        assert checkPatchDivisibility(parameters["patch_size"]) == True, (
            "The patch size is not divisible by 16, which is required for "
            + parameters["model"]["architecture"]
        )

        assert parameters["model"]["base_filters"] % 4 == 0, (
            "The 'base_filters' should be divisible by '4'"
            + " for "
            + parameters["model"]["architecture"]
        )

        self.conv0_1x1 = IncConv(
            self.n_channels,
            self.base_filters,
            self.Conv,
            self.Dropout,
            self.InstanceNorm,
        )
        self.rn_0 = ResNetModule(
            self.base_filters,
            self.base_filters,
            self.Conv,
            self.Dropout,
            self.InstanceNorm,
            res=True,
        )
        self.ri_0 = InceptionModule(
            self.base_filters,
            self.base_filters,
            self.Conv,
            self.Dropout,
            self.InstanceNorm,
            res=True,
        )
        self.ds_0 = IncDownsamplingModule(
            self.base_filters,
            self.base_filters * 2,
            self.Conv,
            self.Dropout,
            self.InstanceNorm,
        )
        self.ri_1 = InceptionModule(
            self.base_filters * 2,
            self.base_filters * 2,
            self.Conv,
            self.Dropout,
            self.InstanceNorm,
            res=True,
        )
        self.ds_1 = IncDownsamplingModule(
            self.base_filters * 2,
            self.base_filters * 4,
            self.Conv,
            self.Dropout,
            self.InstanceNorm,
        )
        self.ri_2 = InceptionModule(
            self.base_filters * 4,
            self.base_filters * 4,
            self.Conv,
            self.Dropout,
            self.InstanceNorm,
            res=True,
        )
        self.ds_2 = IncDownsamplingModule(
            self.base_filters * 4,
            self.base_filters * 8,
            self.Conv,
            self.Dropout,
            self.InstanceNorm,
        )
        self.ri_3 = InceptionModule(
            self.base_filters * 8,
            self.base_filters * 8,
            self.Conv,
            self.Dropout,
            self.InstanceNorm,
            res=True,
        )
        self.ds_3 = IncDownsamplingModule(
            self.base_filters * 8,
            self.base_filters * 16,
            self.Conv,
            self.Dropout,
            self.InstanceNorm,
        )
        self.ri_4 = InceptionModule(
            self.base_filters * 16,
            self.base_filters * 16,
            self.Conv,
            self.Dropout,
            self.InstanceNorm,
            res=True,
        )
        self.us_3 = IncUpsamplingModule(
            self.base_filters * 16,
            self.base_filters * 8,
            self.ConvTranspose,
            self.Dropout,
            self.InstanceNorm,
        )
        self.ri_5 = InceptionModule(
            self.base_filters * 16,
            self.base_filters * 16,
            self.Conv,
            self.Dropout,
            self.InstanceNorm,
            res=True,
        )
        self.us_2 = IncUpsamplingModule(
            self.base_filters * 16,
            self.base_filters * 4,
            self.ConvTranspose,
            self.Dropout,
            self.InstanceNorm,
        )
        self.ri_6 = InceptionModule(
            self.base_filters * 8,
            self.base_filters * 8,
            self.Conv,
            self.Dropout,
            self.InstanceNorm,
            res=True,
        )
        self.us_1 = IncUpsamplingModule(
            self.base_filters * 8,
            self.base_filters * 2,
            self.ConvTranspose,
            self.Dropout,
            self.InstanceNorm,
        )
        self.ri_7 = InceptionModule(
            self.base_filters * 4,
            self.base_filters * 4,
            self.Conv,
            self.Dropout,
            self.InstanceNorm,
            res=True,
        )
        self.us_0 = IncUpsamplingModule(
            self.base_filters * 4,
            self.base_filters,
            self.ConvTranspose,
            self.Dropout,
            self.InstanceNorm,
        )
        self.ri_8 = InceptionModule(
            self.base_filters * 2,
            self.base_filters * 2,
            self.Conv,
            self.Dropout,
            self.InstanceNorm,
            res=True,
        )
        self.conv9_1x1 = IncConv(
            self.base_filters * 2,
            self.base_filters,
            self.Conv,
            self.Dropout,
            self.InstanceNorm,
        )
        self.rn_10 = ResNetModule(
            self.base_filters * 2,
            self.base_filters * 2,
            self.Conv,
            self.Dropout,
            self.InstanceNorm,
            res=True,
        )
        self.dropout = IncDropout(
            self.base_filters * 2,
            self.n_classes,
            self.Conv,
            self.Dropout,
            self.InstanceNorm,
        )

    def forward(self, x):
        """

        Parameters
        ----------
        x : TYPE
            DESCRIPTION.

        Returns
        -------
        x6 : TYPE
            DESCRIPTION.

        """
        x = self.conv0_1x1(x)
        x1 = self.rn_0(x)
        x2 = self.ri_0(x1)
        x3 = self.ds_0(x2)
        x3 = self.ri_1(x3)
        x4 = self.ds_1(x3)
        x4 = self.ri_2(x4)
        x5 = self.ds_2(x4)
        x5 = self.ri_3(x5)
        x6 = self.ds_3(x5)
        x6 = self.ri_4(x6)
        x6 = self.us_3(x6)
        x6 = self.ri_5(torch.cat((x5, x6), dim=1))
        x6 = self.us_2(x6)
        x6 = self.ri_6(torch.cat((x4, x6), dim=1))
        x6 = self.us_1(x6)
        x6 = self.ri_7(torch.cat((x3, x6), dim=1))
        x6 = self.us_0(x6)
        x6 = self.ri_8(torch.cat((x2, x6), dim=1))
        x6 = self.conv9_1x1(x6)
        x6 = self.rn_10(torch.cat((x1, x6), dim=1))
        x6 = self.dropout(x6)

        if not self.final_convolution_layer is None:
            if self.final_convolution_layer == F.softmax:
                x6 = self.final_convolution_layer(x6, dim=1)
            else:
                x6 = self.final_convolution_layer(x6)

        return x6



import yaml
import pickle
import os


def recover_config(modelDir, outputFile):
    assert os.path.exists(
        modelDir
    ), "The model directory does not appear to exist. Please check parameters."

    pickle_location = os.path.join(modelDir, "parameters.pkl")
    assert os.path.exists(
        pickle_location
    ), "The model does not appear to have a configuration file. Please check parameters."

    with open(pickle_location, "rb") as handle:
        parameters = pickle.load(handle)
        os.makedirs(os.path.dirname(outputFile), exist_ok=True)

        # Remove a few problematic objects from the output
        # These cannot be safe_dumped to YAML (or present other problems).
        # To avoid this, try to use primitives and don't use integers as dict keys.
        removable_entries = [
            "output_dir",
            "model_dir_embedded",
            "training_data",
            "validation_data",
            "testing_data",
            "device",
            "subject_spacing",
            "weights",
            "class_weights",
        ]

        for entry in removable_entries:
            parameters.pop(entry, None)

        with open(outputFile, "w") as f:
            print(parameters)
            f.write(
                yaml.safe_dump(parameters, sort_keys=False, default_flow_style=False)
            )

    print(f"Config written to {outputFile}.")
    return True



import os, pathlib, sys, math
import numpy as np
import SimpleITK as sitk
import torchio

from .generic import get_filename_extension_sanitized


def resample_image(
    img, spacing, size=None, interpolator=sitk.sitkLinear, outsideValue=0
):
    """
    Resample image to certain spacing and size.
    Args:
        img (SimpleITK.Image): The input image to resample.
        spacing (list): List of length 3 indicating the voxel spacing as [x, y, z].
        size (list, optional): List of length 3 indicating the number of voxels per dim [x, y, z], which will use compute the appropriate size based on the spacing. Defaults to [].
        interpolator (SimpleITK.InterpolatorEnum, optional): The interpolation type to use. Defaults to SimpleITK.sitkLinear.
        origin (list, optional): The location in physical space representing the [0,0,0] voxel in the input image.  Defaults to [0,0,0].
        outsideValue (int, optional): value used to pad are outside image.  Defaults to 0.
    Raises:
        Exception: Spacing/resolution mismatch.
        Exception: Size mismatch.
    Returns:
        SimpleITK.Image: The resampled input image.
    """
    if len(spacing) != img.GetDimension():
        raise Exception("len(spacing) != " + str(img.GetDimension()))

    # Set Size
    if size is None:
        inSpacing = img.GetSpacing()
        inSize = img.GetSize()
        size = [
            int(math.ceil(inSize[i] * (inSpacing[i] / spacing[i])))
            for i in range(img.GetDimension())
        ]
    else:
        if len(size) != img.GetDimension():
            raise Exception("len(size) != " + str(img.GetDimension()))

    # Resample input image
    return sitk.Resample(
        img,
        size,
        sitk.Transform(),
        interpolator,
        img.GetOrigin(),
        spacing,
        img.GetDirection(),
        outsideValue,
    )


def resize_image(input_image, output_size, interpolator=sitk.sitkLinear):
    """
    This function resizes the input image based on the output size and interpolator.
    Args:
        input_image (SimpleITK.Image): The input image to be resized.
        output_size (numpy.array | list): The output size to resample input_image to.
        interpolator (SimpleITK.sitkInterpolator): The desired interpolator.
    Returns:
        SimpleITK.Image: The output image after resizing.
    """
    output_size_parsed = None
    inputSize = input_image.GetSize()
    inputSpacing = np.array(input_image.GetSpacing())
    outputSpacing = np.array(inputSpacing)

    if isinstance(output_size, dict):
        if "resize" in output_size:
            output_size_parsed = output_size["resize"]
    elif isinstance(output_size, list) or isinstance(output_size, np.array):
        output_size_parsed = output_size

    if len(output_size_parsed) != len(inputSpacing):
        sys.exit(
            "The output size dimension is inconsistent with the input dataset, please check parameters."
        )

    for i, n in enumerate(output_size_parsed):
        outputSpacing[i] = outputSpacing[i] * (inputSize[i] / n)

    return resample_image(
        input_image,
        outputSpacing,
        interpolator=interpolator,
    )


def softer_sanity_check(base_property, new_property, threshold=0.00001):
    """
    This function checks if the new property is within the threshold of the base property.
    Args:
        base_property (float): The base property to check.
        new_property (float): The new property to check
        threshold (float, optional): The threshold to check if the new property is within the base property. Defaults to 0.00001.
    Returns:
        bool: Whether the new property is within the threshold of the base property.
    """
    arr_1 = np.array(base_property)
    arr_2 = np.array(new_property)
    diff = np.sum(arr_1 - arr_2)

    result = False
    if diff <= threshold:
        result = True

    return result


def perform_sanity_check_on_subject(subject, parameters):
    """
    This function performs sanity check on the subject to ensure presence of consistent header information WITHOUT loading images into memory.

    Args:
        subject (torchio.Subject): The input subject.
        parameters (dict): The parameters passed by the user yaml.

    Returns:
        bool: True if everything is okay.

    Raises:
        ValueError: Dimension mismatch in the images.
        ValueError: Origin mismatch in the images.
        ValueError: Orientation mismatch in the images.
    """
    # read the first image and save that for comparison
    file_reader_base = None

    import copy

    list_for_comparison = copy.deepcopy(parameters["headers"]["channelHeaders"])
    if parameters["headers"]["labelHeader"] is not None:
        list_for_comparison.append("label")

    if len(list_for_comparison) > 1:
        for key in list_for_comparison:
            if file_reader_base is None:
                if subject[str(key)]["path"] != "":
                    file_reader_base = sitk.ImageFileReader()
                    file_reader_base.SetFileName(subject[str(key)]["path"])
                    file_reader_base.ReadImageInformation()
                else:
                    # this case is required if any tensor/imaging operation has been applied in dataloader
                    file_reader_base = subject[str(key)].as_sitk()
            else:
                if subject[str(key)]["path"] != "":
                    # in this case, file_reader_base is ready
                    file_reader_current = sitk.ImageFileReader()
                    file_reader_current.SetFileName(subject[str(key)]["path"])
                    file_reader_current.ReadImageInformation()
                else:
                    # this case is required if any tensor/imaging operation has been applied in dataloader
                    file_reader_current = subject[str(key)].as_sitk()

                # this check needs to be absolute
                if (
                    file_reader_base.GetDimension()
                    != file_reader_current.GetDimension()
                ):
                    raise ValueError(
                        "Dimensions for Subject '"
                        + subject["subject_id"]
                        + "' are not consistent."
                    )

                # other checks can be softer
                if not softer_sanity_check(
                    file_reader_base.GetOrigin(), file_reader_current.GetOrigin()
                ):
                    raise ValueError(
                        "Origin for Subject '"
                        + subject["subject_id"]
                        + "' are not consistent."
                    )

                if not softer_sanity_check(
                    file_reader_base.GetDirection(), file_reader_current.GetDirection()
                ):
                    raise ValueError(
                        "Orientation for Subject '"
                        + subject["subject_id"]
                        + "' are not consistent."
                    )

                if not softer_sanity_check(
                    file_reader_base.GetSpacing(), file_reader_current.GetSpacing()
                ):
                    raise ValueError(
                        "Spacing for Subject '"
                        + subject["subject_id"]
                        + "' are not consistent."
                    )

    return True


def write_training_patches(subject, params):
    """
    This function writes the training patches to disk.

    Args:
        subject (torchio.Subject): The input subject.
        params (dict): The parameters passed by the user yaml; needs the "output_dir" and "current_epoch" keys to be present.
    """
    # create folder tree for saving the patches
    training_output_dir = os.path.join(params["output_dir"], "training_patches")
    pathlib.Path(training_output_dir).mkdir(parents=True, exist_ok=True)
    training_output_dir_epoch = os.path.join(
        training_output_dir, str(params["current_epoch"])
    )
    pathlib.Path(training_output_dir_epoch).mkdir(parents=True, exist_ok=True)
    training_output_dir_current_subject = os.path.join(
        training_output_dir_epoch, subject["subject_id"][0]
    )
    pathlib.Path(training_output_dir_current_subject).mkdir(parents=True, exist_ok=True)

    # write the training patches to disk
    ext = get_filename_extension_sanitized(subject["path_to_metadata"][0])
    for key in params["channel_keys"]:
        img_to_write = torchio.ScalarImage(
            tensor=subject[key][torchio.DATA][0], affine=subject[key]["affine"][0]
        ).as_sitk()
        sitk.WriteImage(
            img_to_write,
            os.path.join(training_output_dir_current_subject, "modality_" + key + ext),
        )

    if params["label_keys"] is not None:
        img_to_write = torchio.ScalarImage(
            tensor=subject[params["label_keys"][0]][torchio.DATA][0],
            affine=subject[key]["affine"][0],
        ).as_sitk()
        sitk.WriteImage(
            img_to_write,
            os.path.join(training_output_dir_current_subject, "label_" + key + ext),
        )



from torch.optim import (
    SGD,
    ASGD,
    Rprop,
    Adam,
    AdamW,
    # SparseAdam,
    Adamax,
    Adadelta,
    Adagrad,
    RMSprop,
)


def sgd(parameters):
    # pick defaults
    if not ("momentum" in parameters["optimizer"]):
        parameters["optimizer"]["momentum"] = 0.9
    if not ("weight_decay" in parameters["optimizer"]):
        parameters["optimizer"]["weight_decay"] = 0
    if not ("dampening" in parameters["optimizer"]):
        parameters["optimizer"]["dampening"] = 0
    if not ("nesterov" in parameters["optimizer"]):
        parameters["optimizer"]["nesterov"] = False
    return SGD(
        parameters["model_parameters"],
        lr=parameters["learning_rate"],
        momentum=parameters["optimizer"]["momentum"],
        weight_decay=parameters["optimizer"]["weight_decay"],
        dampening=parameters["optimizer"]["dampening"],
        nesterov=parameters["optimizer"]["nesterov"],
    )


def asgd(parameters):
    # pick defaults
    if not ("lambd" in parameters["optimizer"]):
        parameters["optimizer"]["lambd"] = 1e-4
    if not ("alpha" in parameters["optimizer"]):
        parameters["optimizer"]["alpha"] = 0.75
    if not ("t0" in parameters["optimizer"]):
        parameters["optimizer"]["t0"] = 1e6
    if not ("weight_decay" in parameters["optimizer"]):
        parameters["optimizer"]["weight_decay"] = 0
    return ASGD(
        parameters["model_parameters"],
        lr=parameters["learning_rate"],
        alpha=parameters["optimizer"]["alpha"],
        t0=parameters["optimizer"]["t0"],
        lambd=parameters["optimizer"]["lambd"],
        weight_decay=parameters["optimizer"]["weight_decay"],
    )


def adam(parameters, opt_type="normal"):
    # pick defaults
    if not ("betas" in parameters["optimizer"]):
        parameters["optimizer"]["betas"] = (0.9, 0.999)
    if not ("weight_decay" in parameters["optimizer"]):
        parameters["optimizer"]["weight_decay"] = 0.00005
    if not ("eps" in parameters["optimizer"]):
        parameters["optimizer"]["eps"] = 1e-8
    if not ("amsgrad" in parameters["optimizer"]):
        parameters["optimizer"]["amsgrad"] = False

    if opt_type == "normal":
        function = Adam
    else:
        function = AdamW
    return function(
        parameters["model_parameters"],
        lr=parameters["learning_rate"],
        betas=parameters["optimizer"]["betas"],
        weight_decay=parameters["optimizer"]["weight_decay"],
        eps=parameters["optimizer"]["eps"],
        amsgrad=parameters["optimizer"]["amsgrad"],
    )


def adamw(parameters):
    return adam(parameters, opt_type="adamw")


def adamax(parameters):
    # pick defaults
    if not ("betas" in parameters["optimizer"]):
        parameters["optimizer"]["betas"] = (0.9, 0.999)
    if not ("weight_decay" in parameters["optimizer"]):
        parameters["optimizer"]["weight_decay"] = 0.00005
    if not ("eps" in parameters["optimizer"]):
        parameters["optimizer"]["eps"] = 1e-8

    return Adamax(
        parameters["model_parameters"],
        lr=parameters["learning_rate"],
        betas=parameters["optimizer"]["betas"],
        weight_decay=parameters["optimizer"]["weight_decay"],
        eps=parameters["optimizer"]["eps"],
    )


# def sparseadam(parameters):
#     # pick defaults
#     if not ("betas" in parameters["optimizer"]):
#         parameters["optimizer"]["betas"] = (0.9, 0.999)
#     if not ("eps" in parameters["optimizer"]):
#         parameters["optimizer"]["eps"] = 1e-8

#     return SparseAdam(
#         parameters["model_parameters"],
#         lr=parameters["learning_rate"],
#         betas=parameters["optimizer"]["betas"],
#         eps=parameters["optimizer"]["eps"],
#     )


def rprop(parameters):
    if not ("etas" in parameters["optimizer"]):
        parameters["optimizer"]["etas"] = (0.5, 1.2)
    if not ("step_sizes" in parameters["optimizer"]):
        parameters["optimizer"]["step_sizes"] = (1e-7, 50)
    return Rprop(
        parameters["model_parameters"],
        lr=parameters["learning_rate"],
        etas=parameters["optimizer"]["etas"],
        step_sizes=parameters["optimizer"]["step_sizes"],
    )


def adadelta(parameters):
    # pick defaults
    if not ("rho" in parameters["optimizer"]):
        parameters["optimizer"]["rho"] = 0.9
    if not ("eps" in parameters["optimizer"]):
        parameters["optimizer"]["eps"] = 1e-6
    if not ("weight_decay" in parameters["optimizer"]):
        parameters["optimizer"]["weight_decay"] = 0
    return Adadelta(
        parameters["model_parameters"],
        lr=parameters["learning_rate"],
        rho=parameters["optimizer"]["rho"],
        eps=parameters["optimizer"]["eps"],
        weight_decay=parameters["optimizer"]["weight_decay"],
    )


def adagrad(parameters):
    # pick defaults
    if not ("lr_decay" in parameters["optimizer"]):
        parameters["optimizer"]["lr_decay"] = 0
    if not ("eps" in parameters["optimizer"]):
        parameters["optimizer"]["eps"] = 1e-6
    if not ("weight_decay" in parameters["optimizer"]):
        parameters["optimizer"]["weight_decay"] = 0

    return Adagrad(
        parameters["model_parameters"],
        lr=parameters["learning_rate"],
        lr_decay=parameters["optimizer"]["lr_decay"],
        eps=parameters["optimizer"]["eps"],
        weight_decay=parameters["optimizer"]["weight_decay"],
    )


def rmsprop(parameters):
    # pick defaults
    if not ("momentum" in parameters["optimizer"]):
        parameters["optimizer"]["momentum"] = 0
    if not ("weight_decay" in parameters["optimizer"]):
        parameters["optimizer"]["weight_decay"] = 0
    if not ("alpha" in parameters["optimizer"]):
        parameters["optimizer"]["alpha"] = 0.99
    if not ("eps" in parameters["optimizer"]):
        parameters["optimizer"]["eps"] = 1e-8
    if not ("centered" in parameters["optimizer"]):
        parameters["optimizer"]["centered"] = False

    return RMSprop(
        parameters["model_parameters"],
        lr=parameters["learning_rate"],
        alpha=parameters["optimizer"]["alpha"],
        eps=parameters["optimizer"]["eps"],
        centered=parameters["optimizer"]["centered"],
        momentum=parameters["optimizer"]["momentum"],
        weight_decay=parameters["optimizer"]["weight_decay"],
    )



# adapted from https://github.com/fepegar/torchio/blob/main/src/torchio/transforms/augmentation/intensity/random_blur.py

from collections import defaultdict
from typing import Union, Tuple

import torch
import numpy as np
import scipy.ndimage as ndi

from torchio.typing import TypeData, TypeTripletFloat, TypeSextetFloat
from torchio.data.subject import Subject
from torchio.transforms import IntensityTransform
from torchio.transforms.augmentation import RandomTransform
from torchio.transforms.augmentation.intensity.random_blur import Blur


def blur(
    data: TypeData,
    spacing: TypeTripletFloat,
    std_voxel: TypeTripletFloat,
) -> torch.Tensor:
    assert data.ndim == 3
    std_physical = np.array(std_voxel) / np.array(spacing)
    blurred = ndi.gaussian_filter(data, std_physical)
    tensor = torch.as_tensor(blurred)
    return tensor


class RandomBlurEnhanced(RandomTransform, IntensityTransform):
    r"""Blur an image using a random-sized Gaussian filter.

    Args:
        std: Tuple :math:`(a_1, b_1, a_2, b_2, a_3, b_3)` representing the
            ranges (in mm) of the standard deviations
            :math:`(\sigma_1, \sigma_2, \sigma_3)` of the Gaussian kernels used
            to blur the image along each axis, where
            :math:`\sigma_i \sim \mathcal{U}(a_i, b_i)`.
            If two values :math:`(a, b)` are provided,
            then :math:`\sigma_i \sim \mathcal{U}(a, b)`.
            If only one value :math:`x` is provided,
            then :math:`\sigma_i \sim \mathcal{U}(0, x)`.
            If three values :math:`(x_1, x_2, x_3)` are provided,
            then :math:`\sigma_i \sim \mathcal{U}(0, x_i)`.
        **kwargs: See :class:`~torchio.transforms.Transform` for additional
            keyword arguments.
    """

    def __init__(self, std: Union[float, Tuple[float, float]] = None, **kwargs):
        super().__init__(**kwargs)
        self.std_original = std

    def apply_transform(self, subject: Subject) -> Subject:
        arguments = defaultdict(dict)
        for name, image in self.get_images_dict(subject).items():
            self.std_ranges = self.calculate_std_ranges(image)
            arguments["std"][name] = self.get_params(self.std_ranges)
        transform = Blur(**self.add_include_exclude(arguments))
        transformed = transform(subject)
        return transformed

    def get_params(self, std_ranges: TypeSextetFloat) -> TypeTripletFloat:
        std = self.sample_uniform_sextet(std_ranges)
        return std

    def calculate_std_ranges(self, image: torch.Tensor) -> tuple:
        std_ranges = self.std_original
        if self.std_original is None:
            # calculate the default std range based on 1.5% of the input image std - https://github.com/mlcommons/GaNDLF/issues/518
            std_ranges = (0, 0.015 * torch.std(image.data).item())
        return self.parse_params(std_ranges, None, "std", min_constraint=0)



import sys
import torch.nn as nn
import torch.nn.functional as F
from collections import OrderedDict
import numpy as np

from .modelBase import ModelBase


class ResNet(ModelBase):
    """
    Initializer function for the Resnet model

    Args:
        configuration (dict): A dictionary of configuration parameters for the model.
        parameters (dict) - overall parameters dictionary
    """

    def __init__(
        self,
        parameters: dict,
        blockType,  # basic block or bottleneck
        block_config,
    ):
        super(ResNet, self).__init__(parameters)

        allowedLay = checkPatchDimensions(parameters["patch_size"], len(block_config))

        if allowedLay != len(block_config) and allowedLay >= 1:
            print(
                "The patch size is not large enough for desired number of layers.",
                " It is expected that each dimension of the patch size is 2^(layers + 1)*i, where i is in a integer greater than 2.",
                "Only the first %d layers will run." % allowedLay,
            )

        elif allowedLay != len(block_config) and allowedLay <= 0:
            sys.exit(
                "The patch size is not large enough for desired number of layers.",
                " It is expected that each dimension of the patch size is 2^(layers + 1)*i, where i is in a integer greater than 2.",
            )

        block_config = block_config[:allowedLay]

        # check/define defaults
        if not ("num_init_features" in parameters):
            parameters["num_init_features"] = 64
        if self.n_dimensions == 2:
            self.output_size = (1, 1)
        elif self.n_dimensions == 3:
            self.output_size = (1, 1, 1)
        else:
            sys.exit("Only 2D or 3D convolutions are supported.")
        if self.Norm is None:
            sys.stderr.write(
                "Warning: resnet is not defined without a normalization layer"
            )
            self.Norm = self.BatchNorm

        # first convolution: 7x7 conv stride 2, 2x2 pool stride 2
        self.features = [
            (
                "conv1",
                self.Conv(
                    in_channels=self.n_channels,
                    out_channels=parameters["num_init_features"],
                    kernel_size=7,
                    stride=2,
                    padding=3,
                    bias=False,
                ),
            ),
            ("norm1", self.Norm(parameters["num_init_features"])),
            ("relu1", nn.ReLU(inplace=True)),
            ("pool1", self.MaxPool(kernel_size=3, stride=2, padding=1)),
        ]
        self.features = nn.Sequential(OrderedDict(self.features))

        # make conv blocks
        num_features = parameters["num_init_features"]
        offset = num_features - num_features * 2**-1
        for i, num_lay in enumerate(block_config):
            block = blockType(
                num_in_feats=int(num_features * 2 ** (i - 1) + offset),
                num_out_feats=int(num_features * 2**i),
                num_layers=num_lay,
                Norm=self.Norm,
                Conv=self.Conv,
                num_block=i,
            )
            self.features.add_module("block{}".format(i + 1), block)
            offset = 0

        # final layer, fully connected -> number classes
        if blockType == _BottleNeckBlock:
            self.classifier = nn.Linear(4 * num_features * 2**i, self.n_classes)
        else:
            self.classifier = nn.Linear(num_features * 2**i, self.n_classes)

        # define starting weights
        for m in self.modules():
            if isinstance(m, self.Conv):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, self.Norm):
                if m.weight is not None:
                    m.weight.data.fill_(1)
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        features = self.features(x)
        out = self.AdaptiveAvgPool(self.output_size)(features).view(
            features.size(0), -1
        )
        out = self.classifier(out)

        if not self.final_convolution_layer is None:
            if self.final_convolution_layer == F.softmax:
                out = self.final_convolution_layer(out, dim=1)
            else:
                out = self.final_convolution_layer(out)

        return out


class _BasicBlock(nn.Sequential):
    def __init__(self, num_in_feats, num_out_feats, num_layers, Norm, Conv, num_block):
        super().__init__()

        # iterate through size of block
        num_feats = num_in_feats
        for i_lay in range(0, num_layers):
            # add basic layer
            layer = _BasicLayer(
                num_in_feats=num_feats,
                num_out_feats=num_out_feats,
                Norm=Norm,
                Conv=Conv,
                downsample=(num_block != 0) and (i_lay == 0),
            )
            self.add_module("layer{}".format(i_lay + 1), layer)
            num_feats = num_out_feats


class _BasicLayer(nn.Sequential):
    def __init__(
        self,
        num_in_feats,
        num_out_feats,
        Norm,
        Conv,
        downsample,
    ):
        super().__init__()

        self.sizing = num_out_feats != num_in_feats  # true if size changes
        if self.sizing:  # add module to project input to correct size
            self.add_module(
                "project",
                Conv(
                    num_in_feats,
                    num_out_feats,
                    kernel_size=1,
                    stride=1 + downsample,
                    padding=0,
                    bias=False,
                ),
            )
            self.add_module("projectNorm", Norm(num_out_feats))

        # 3x3 --> 3x3

        self.add_module(
            "conv1",
            Conv(
                num_in_feats,
                num_out_feats,
                kernel_size=3,
                stride=1 + downsample,
                padding=1,
                bias=False,
            ),
        )
        self.add_module("norm1", Norm(num_out_feats))
        self.add_module("relu", nn.ReLU(inplace=True))

        self.add_module(
            "conv2",
            Conv(
                num_out_feats,
                num_out_feats,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=False,
            ),
        )

        self.add_module("norm2", Norm(num_out_feats))

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.norm1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.norm2(out)

        if self.sizing:  # project input to correct output size if needed
            identity = self.project(identity)
            identity = self.projectNorm(identity)

        out += identity
        out = self.relu(out)
        return out


class _BottleNeckBlock(nn.Sequential):
    def __init__(self, num_in_feats, num_out_feats, num_layers, Norm, Conv, num_block):
        super().__init__()

        # iterate through size of block
        if num_block != 0:
            num_feats = 4 * num_in_feats
        else:
            num_feats = num_in_feats

        for i_lay in range(0, num_layers):
            # add basic layer
            layer = _BottleNeckLayer(
                num_in_feats=num_feats,
                num_out_feats=num_out_feats,
                Norm=Norm,
                Conv=Conv,
                downsample=(num_block != 0) and (i_lay == 0),
            )
            self.add_module("layer{}".format(i_lay + 1), layer)
            num_feats = 4 * num_out_feats


class _BottleNeckLayer(nn.Sequential):
    def __init__(
        self,
        num_in_feats,  # true input number
        num_out_feats,  # 3rd conv increases features --> true output number - 4*num_out_feats
        Norm,
        Conv,
        downsample,
    ):
        super().__init__()

        self.sizing = 4 * num_out_feats != num_in_feats  # true if size changes

        if self.sizing:  # add module to project input to correct size
            self.add_module(
                "project",
                Conv(
                    num_in_feats,
                    4 * num_out_feats,
                    kernel_size=1,
                    stride=1 + downsample,
                    padding=0,
                    bias=False,
                ),
            )
            self.add_module("projectNorm", Norm(4 * num_out_feats))

        # 1x1 conv (k) -> 3x3 conv (k) -> 1x1 conv (4*k)

        self.add_module(
            "conv1",
            Conv(
                num_in_feats,
                num_out_feats,
                kernel_size=1,
                stride=1 + downsample,
                padding=0,
                bias=False,
            ),
        )
        self.add_module("norm1", Norm(num_out_feats))
        self.add_module("relu", nn.ReLU(inplace=True))

        self.add_module(
            "conv2",
            Conv(
                num_out_feats,
                num_out_feats,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=False,
            ),
        )

        self.add_module("norm2", Norm(num_out_feats))

        self.add_module(
            "conv3",
            Conv(
                num_out_feats,
                4 * num_out_feats,
                kernel_size=1,
                stride=1,
                padding=0,
                bias=False,
            ),
        )

        self.add_module("norm3", Norm(4 * num_out_feats))

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.norm1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.norm2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.norm3(out)

        if self.sizing:  # project input to correct output size if needed
            identity = self.project(identity)
            identity = self.projectNorm(identity)

        out += identity
        out = self.relu(out)
        return out


def checkPatchDimensions(patch_size, numlay):
    if isinstance(patch_size, int):
        patch_size_to_check = np.array(patch_size)
    else:
        patch_size_to_check = patch_size
    # for 2D, don't check divisibility of last dimension
    if patch_size_to_check[-1] == 1:
        patch_size_to_check = patch_size_to_check[:-1]

    if all(
        [
            x >= 2 ** (numlay + 2) and x % 2 ** (numlay + 1) == 0
            for x in patch_size_to_check
        ]
    ):
        return numlay
    else:
        # base2 = np.floor(np.log2(patch_size_to_check))
        base2 = np.array([getBase2(x) for x in patch_size_to_check])
        remain = patch_size_to_check / 2**base2  # check that at least 1

        layers = np.where(remain == 1, base2 - 1, base2)
        return int(np.min(layers) - 1)


def getBase2(num):
    base = 0
    while num % 2 == 0:
        num = num / 2
        base = base + 1
    return base


def resnet18(parameters):
    return ResNet(parameters, _BasicBlock, block_config=(2, 2, 2, 2))


def resnet34(parameters):
    return ResNet(parameters, _BasicBlock, block_config=(3, 4, 6, 3))


def resnet50(parameters):
    return ResNet(parameters, _BottleNeckBlock, block_config=(3, 4, 6, 3))


def resnet101(parameters):
    return ResNet(parameters, _BottleNeckBlock, block_config=(3, 4, 23, 3))


def resnet152(parameters):
    return ResNet(parameters, _BottleNeckBlock, block_config=(3, 8, 36, 3))



import typing, sys
import torch
import torch.nn as nn
import torch.nn.functional as F

from GANDLF.models.seg_modules.add_conv_block import add_conv_block
from GANDLF.models.seg_modules.add_downsample_conv_block import (
    add_downsample_conv_block,
)
from GANDLF.models.unet import unet
from .modelBase import ModelBase
from copy import deepcopy


class Decoder(ModelBase):
    def __init__(
        self,
        parameters,
        anatomy_factors,
        num_layers=5,
    ):
        super(Decoder, self).__init__(parameters)
        self.num_layers = num_layers
        self.layer_list = add_conv_block(
            self.Conv, self.BatchNorm, in_ch=anatomy_factors, out_ch=self.base_filters
        )
        for _ in range(self.num_layers - 2):
            self.layer_list += add_conv_block(
                self.Conv,
                self.BatchNorm,
                in_ch=self.base_filters,
                out_ch=self.base_filters,
            )
        self.conv = self.Conv(self.base_filters, 1, 3, 1, 1)
        # Add layers to Module List
        self.layers = nn.ModuleList(self.layer_list)
        self.apply(self.weight_init)

        nn.init.xavier_normal_(self.conv.weight.data)
        self.conv.bias.data.zero_()

    @staticmethod
    def CalcVectorMeanStd(feat, eps=1e-5):
        # eps is a small value added to the variance to avoid divide-by-zero.
        feat_var = feat.var(dim=1) + eps
        feat_std = feat_var.sqrt()
        feat_mean = feat.mean(dim=1)
        return feat_mean, feat_std

    @staticmethod
    def CalcTensorMeanStd(feat, eps=1e-5):
        # eps is a small value added to the variance to avoid divide-by-zero.
        size = feat.size()
        assert len(size) == 4
        N, C = size[:2]
        feat_var = feat.view(N, C, -1).var(dim=2) + eps
        feat_std = feat_var.sqrt().view(N, C, 1, 1)
        feat_mean = feat.view(N, C, -1).mean(dim=2).view(N, C, 1, 1)
        return feat_mean, feat_std

    @staticmethod
    def weight_init(model):
        for module in model.modules():
            if isinstance(module, nn.Conv2d) or isinstance(module, nn.Linear):
                nn.init.kaiming_normal_(module.weight)
                if module.bias is not None:
                    module.bias.data.zero_()

    @staticmethod
    def AdaIN(content_feat, style_feat):
        size = content_feat.size()
        style_mean, style_std = Decoder.CalcVectorMeanStd(style_feat)
        content_mean, content_std = Decoder.CalcTensorMeanStd(content_feat)
        normalized_feat = (
            content_feat - content_mean.expand(size)
        ) / content_std.expand(size)
        return normalized_feat * style_std.view(style_std.shape[0], 1, 1, 1).expand(
            size
        ) + style_mean.view(style_mean.shape[0], 1, 1, 1).expand(size)

    def forward(self, c: torch.Tensor, s: torch.Tensor) -> torch.Tensor:
        x = c
        for i, f in enumerate(self.layers):
            if (i % 2) == 0:
                x = f(x)
            else:
                x = F.leaky_relu(f(x), 0.2)
                x = Decoder.AdaIN(x, s)
        x = torch.tanh(self.conv(x))
        return x


class Segmentor(ModelBase):
    def __init__(
        self,
        parameters,
        anatomy_factors,
    ):
        super(Segmentor, self).__init__(parameters)
        self.layer_list = add_conv_block(
            self.Conv,
            self.BatchNorm,
            in_ch=anatomy_factors,
            out_ch=self.base_filters * 4,
        )
        self.layer_list += add_conv_block(
            self.Conv,
            self.BatchNorm,
            in_ch=self.base_filters * 4,
            out_ch=self.base_filters * 4,
        )
        self.conv = self.Conv(self.base_filters * 4, self.n_classes, 1, 1, 0)
        # Add layers to Module List
        self.layers = nn.ModuleList(self.layer_list)
        self.apply(self.weight_init)

        nn.init.xavier_normal_(self.conv.weight.data)
        self.conv.bias.data.zero_()

    @staticmethod
    def weight_init(model):
        for module in model.modules():
            if isinstance(module, nn.Conv2d) or isinstance(module, nn.Linear):
                nn.init.kaiming_normal_(module.weight)
                if module.bias is not None:
                    module.bias.data.zero_()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        for i, f in enumerate(self.layers):
            if (i % 2) == 0:
                x = f(x)
            else:
                x = F.leaky_relu(f(x), 0.2)
        x = F.softmax(self.conv(x), dim=1)
        return x


class ModalityEncoder(ModelBase):
    def __init__(
        self,
        parameters,
        anatomy_factors,
        modality_factors,
        num_layers=4,
    ):
        super(ModalityEncoder, self).__init__(parameters)
        self.num_layers = num_layers
        self.layer_list = add_downsample_conv_block(
            self.Conv, self.BatchNorm, in_ch=anatomy_factors + 1, out_ch=16
        )
        for _ in range(self.num_layers - 1):
            self.layer_list += add_downsample_conv_block(
                self.Conv, self.BatchNorm, in_ch=16, out_ch=16
            )
        self.fc = nn.Linear(3136, 32)
        self.norm = nn.BatchNorm1d(32)
        self.mu_fc = nn.Linear(32, modality_factors)
        self.logvar_fc = nn.Linear(32, modality_factors)

        # Add layers to Module List
        self.layers = nn.ModuleList(self.layer_list)
        self.apply(self.weight_init)

        nn.init.xavier_normal_(self.fc.weight.data)
        self.fc.bias.data.zero_()
        nn.init.xavier_normal_(self.mu_fc.weight.data)
        self.mu_fc.bias.data.zero_()
        nn.init.xavier_normal_(self.logvar_fc.weight.data)
        self.logvar_fc.bias.data.zero_()

    @staticmethod
    def weight_init(model):
        for module in model.modules():
            if isinstance(module, nn.Conv2d) or isinstance(module, nn.Linear):
                nn.init.kaiming_normal_(module.weight)
                if module.bias is not None:
                    module.bias.data.zero_()

    def forward(self, x: torch.Tensor, c: torch.Tensor) -> torch.Tensor:
        x = torch.cat((x, c), 1)
        for i, f in enumerate(self.layers):
            if (i % 2) == 0:
                x = f(x)
            else:
                x = F.leaky_relu(f(x), 0.2)
        x = self.norm(self.fc(x.view(-1, 3136)))
        x = F.leaky_relu(x, 0.3)
        mu = self.mu_fc(x)
        logvar = self.logvar_fc(x)
        return mu, logvar


class SDNet(ModelBase):
    def __init__(
        self,
        parameters: dict,
    ):
        super(SDNet, self).__init__(parameters)
        self.anatomy_factors = 8
        self.modality_factors = 8

        if parameters["patch_size"] != [224, 224, 1]:
            print(
                "WARNING: The patch size is not 224x224, which is required for sdnet. Using default patch size instead",
                file=sys.stderr,
            )
            parameters["patch_size"] = [224, 224, 1]

        if parameters["batch_size"] == 1:
            raise ValueError("'batch_size' needs to be greater than 1 for 'sdnet'")

        # amp is not supported for sdnet
        parameters["model"]["amp"] = False
        parameters["model"]["norm_type"] = "instance"

        parameters_unet = deepcopy(parameters)
        parameters_unet["model"]["num_classes"] = self.anatomy_factors
        parameters_unet["model"]["norm_type"] = "instance"
        parameters_unet["model"]["final_layer"] = None

        self.cencoder = unet(parameters_unet)
        self.mencoder = ModalityEncoder(
            parameters,
            self.anatomy_factors,
            self.modality_factors,
        )
        self.decoder = Decoder(
            parameters,
            self.anatomy_factors,
        )
        self.segmentor = Segmentor(
            parameters,
            self.anatomy_factors,
        )

    @staticmethod
    def reparameterize(mu: torch.Tensor, logvar: torch.Tensor) -> torch.Tensor:
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        return mu + eps * std

    def forward(self, x: torch.Tensor) -> typing.List[torch.Tensor]:
        if x.shape[1] > 1:
            x = x[:, 0:1, :, :]
        anatomy_factors = F.gumbel_softmax(self.cencoder(x), hard=True, dim=1)
        mu, logvar = self.mencoder(x, anatomy_factors)
        modality_factors = SDNet.reparameterize(mu, logvar)
        sm = self.segmentor(anatomy_factors)
        reco = self.decoder(anatomy_factors, modality_factors)
        modality_factors_reencoded, _ = self.mencoder(reco, anatomy_factors)
        # sm, anatomy_factors, mu, logvar, modality_factors_reencoded
        return (
            sm,
            reco,
            mu,
            logvar,
            modality_factors_reencoded,
        )



# -*- coding: utf-8 -*-
"""
Modified from https://github.com/pytorch/vision.git
"""

import math
import torch.nn as nn
import torch.nn.functional as F

from .modelBase import ModelBase


class VGG(ModelBase):
    """
    VGG model
    """

    def __init__(
        self,
        parameters: dict,
        configuration,
    ):
        """
        Initializer function for the VGG model

        Args:
            configuration (dict): A dictionary of configuration parameters for the model.
            parameters (dict) - overall parameters dictionary; parameters specific for DenseNet:
        """
        super(VGG, self).__init__(parameters)

        self.features = self.make_layers(
            configuration,
            self.n_channels,
        )

        # amp is not supported for vgg
        parameters["model"]["amp"] = False

        self.classifier = nn.Sequential(
            nn.Dropout(),
            self.GlobalAvgPool(),
            nn.ReLU(True),
            nn.Dropout(),
            # number of input features should be changed later, but works for all vgg right now
            nn.Linear(512, self.n_classes),
        )
        # Initialize weights
        for m in self.modules():
            if isinstance(m, self.Conv):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2.0 / n))
                m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight)
            else:
                pass

    def forward(self, x):
        out = self.features(x)
        out = self.classifier(out)
        if not self.final_convolution_layer is None:
            if self.final_convolution_layer == F.softmax:
                out = self.final_convolution_layer(out, dim=1)
            else:
                out = self.final_convolution_layer(out)
        return out

    def make_layers(self, cfg, in_channels):
        layers = []
        for v in cfg:
            if v == "M":
                layers += [self.MaxPool(kernel_size=2, stride=2)]
            else:
                conv = self.Conv(in_channels, v, kernel_size=3, padding=1)
                if self.norm_type in ["batch", "instance"]:
                    layers += [conv, self.Norm(v), nn.ReLU(inplace=True)]
                else:
                    layers += [conv, nn.ReLU(inplace=True)]
                in_channels = v
        return nn.Sequential(*layers)


cfg = {
    "A": [64, "M", 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"],
    "B": [64, 64, "M", 128, 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"],
    "D": [
        64,
        64,
        "M",
        128,
        128,
        "M",
        256,
        256,
        256,
        "M",
        512,
        512,
        512,
        "M",
        512,
        512,
        512,
        "M",
    ],
    "E": [
        64,
        64,
        "M",
        128,
        128,
        "M",
        256,
        256,
        256,
        256,
        "M",
        512,
        512,
        512,
        512,
        "M",
        512,
        512,
        512,
        512,
        "M",
    ],
}


class vgg11(VGG):
    """
    VGG model
    """

    def __init__(
        self,
        parameters,
    ):
        super(vgg11, self).__init__(parameters=parameters, configuration=cfg["A"])


class vgg13(VGG):
    """
    VGG model
    """

    def __init__(
        self,
        parameters,
    ):
        super(vgg13, self).__init__(parameters=parameters, configuration=cfg["B"])


class vgg16(VGG):
    """
    VGG model
    """

    def __init__(
        self,
        parameters,
    ):
        super(vgg16, self).__init__(parameters=parameters, configuration=cfg["D"])


class vgg19(VGG):
    """
    VGG model
    """

    def __init__(
        self,
        parameters,
    ):
        super(vgg19, self).__init__(parameters=parameters, configuration=cfg["E"])



import torch
import torch.nn as nn
import torch.nn.functional as F


class out_conv(nn.Module):
    def __init__(
        self,
        input_channels,
        output_channels,
        conv=nn.Conv2d,
        conv_kwargs=None,
        norm=nn.BatchNorm2d,
        norm_kwargs=None,
        act=nn.LeakyReLU,
        act_kwargs=None,
        network_kwargs=None,
        final_convolution_layer=nn.Sigmoid,
        sigmoid_input_multiplier=1.0,
    ):
        nn.Module.__init__(self)
        if conv_kwargs is None:
            conv_kwargs = {"kernel_size": 3, "stride": 1, "padding": 1, "bias": True}
        if norm_kwargs is None:
            norm_kwargs = {
                "eps": 1e-5,
                "affine": True,
                "momentum": 0.1,
                "track_running_stats": True,
            }
        if act_kwargs is None:
            act_kwargs = {"negative_slope": 1e-2, "inplace": True}
        if network_kwargs is None:
            network_kwargs = {"res": False}

        self.residual = network_kwargs["res"]
        self.sigmoid_input_multiplier = sigmoid_input_multiplier

        self.conv0 = conv(input_channels, output_channels, **conv_kwargs)

        self.in_0 = (
            norm(input_channels, **norm_kwargs) if norm is not None else nn.Identity()
        )

        self.act = act(**act_kwargs)

        self.final_convolution_layer = final_convolution_layer

    def forward(self, x):
        x = self.conv0(self.act(self.in_0(x)))

        if not (self.final_convolution_layer is None):
            if self.final_convolution_layer == F.softmax:
                x = self.final_convolution_layer(x, dim=1)
            elif self.final_convolution_layer == torch.sigmoid:
                x = torch.sigmoid(self.sigmoid_input_multiplier * x)
            else:
                x = self.final_convolution_layer(x)
        return x



from .modelBase import ModelBase
from GANDLF.models.seg_modules.out_conv import out_conv
import torch
import torch.nn as nn
from torch.nn import ModuleList
import numpy as np
import math


class _DeconvConvBlock(nn.Sequential):
    def __init__(self, in_feats, out_feats, Norm, Conv, Deconv):
        super().__init__()

        self.add_module(
            "deconv",
            Deconv(
                in_channels=in_feats,
                out_channels=out_feats,
                kernel_size=2,
                stride=2,
                padding=0,
                output_padding=0,
            ),
        )

        self.add_module(
            "conv",
            Conv(
                in_channels=out_feats,
                out_channels=out_feats,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=False,
            ),
        )

        self.add_module("norm", Norm(out_feats))
        self.add_module("relu", nn.ReLU(inplace=True))

    def forward(self, x):
        x = self.deconv(x)
        x = self.conv(x)
        x = self.norm(x)
        x = self.relu(x)

        return x


class _ConvBlock(nn.Sequential):
    def __init__(self, in_feats, out_feats, Norm, Conv):
        super().__init__()

        self.add_module(
            "conv",
            Conv(
                in_channels=in_feats,
                out_channels=out_feats,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=False,
            ),
        )

        self.add_module("norm", Norm(out_feats))
        self.add_module("relu", nn.ReLU(inplace=True))

    def forward(self, x):
        x = self.conv(x)
        x = self.norm(x)
        x = self.relu(x)

        return x


class _UpsampleBlock(nn.Sequential):
    def __init__(self, in_feats, Norm, Conv, Deconv):
        super().__init__()

        self.add_module(
            "conv1",
            Conv(
                in_channels=in_feats,
                out_channels=in_feats // 2,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=False,
            ),
        )

        self.add_module("norm1", Norm(in_feats // 2))
        self.add_module("relu1", nn.ReLU(inplace=True))

        self.add_module(
            "conv2",
            Conv(
                in_channels=in_feats // 2,
                out_channels=in_feats // 2,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=False,
            ),
        )

        self.add_module("norm2", Norm(in_feats // 2))
        self.add_module("relu2", nn.ReLU(inplace=True))

        self.add_module(
            "deconv",
            Deconv(
                in_channels=in_feats // 2,
                out_channels=in_feats // 4,
                kernel_size=2,
                stride=2,
                padding=0,
                output_padding=0,
            ),
        )

    def forward(self, x):
        x = self.conv1(x)
        x = self.norm1(x)
        x = self.relu1(x)

        x = self.conv2(x)
        x = self.norm2(x)
        x = self.relu2(x)

        x = self.deconv(x)

        return x


class _MLP(nn.Sequential):
    def __init__(self, in_feats, out_feats, Norm):
        super().__init__()

        self.add_module("norm", nn.LayerNorm(in_feats))

        self.add_module("linear1", nn.Linear(in_feats, out_feats))
        self.add_module("gelu1", nn.GELU())

        self.add_module("linear2", nn.Linear(out_feats, in_feats))
        self.add_module("gelu2", nn.GELU())

    def forward(self, x):
        x = self.norm(x)
        x = self.linear1(x)
        x = self.gelu1(x)

        x = self.linear2(x)
        x = self.gelu2(x)

        return x


class _MSA(nn.Module):
    def __init__(self, embed_size, num_heads):
        super().__init__()

        self.num_heads = num_heads
        self.query_size = int(embed_size / self.num_heads)
        self.all_heads = self.query_size * self.num_heads  # should equal embed_size

        self.query = nn.Linear(embed_size, self.all_heads)
        self.key = nn.Linear(embed_size, self.all_heads)
        self.value = nn.Linear(embed_size, self.all_heads)

        self.out = nn.Linear(self.all_heads, self.all_heads)

        self.softmax = nn.Softmax(dim=-1)

    def reshape(self, x):
        x_shape = list(x.size()[:-1]) + [self.num_heads, self.query_size]
        x = x.view(*x_shape)
        return x.permute(0, 2, 1, 3)

    def forward(self, x):
        query = self.reshape(self.query(x))
        key = self.reshape(self.key(x))
        value = self.reshape(self.value(x))

        attention_weights = self.softmax(
            torch.matmul(query, key.transpose(-1, -2)) / math.sqrt(self.query_size)
        )
        self_attention = torch.matmul(attention_weights, value)
        self_attention = self_attention.permute(0, 2, 1, 3).contiguous()
        self_attention = self_attention.view(
            list(self_attention.size()[:-2]) + [self.all_heads]
        )  # reshape

        msa = self.out(self_attention)

        return msa


class _Embedding(nn.Module):
    def __init__(self, img_size, patch_size, in_feats, embed_size, Conv):
        super().__init__()
        # self.n_patches = int(np.prod(img_size) / (patch_size**3))
        self.n_patches = int(np.prod([i / patch_size for i in img_size]))

        self.pos_embed = nn.Parameter(torch.zeros(1, self.n_patches, embed_size))
        self.patch_embed = Conv(
            in_channels=in_feats,
            out_channels=embed_size,
            kernel_size=patch_size,
            stride=patch_size,
        )

    def forward(self, x):
        x = self.patch_embed(x)
        x = x.flatten(2)
        x = x.transpose(-1, -2)

        x = x + self.pos_embed

        return x


class _TransformerLayer(nn.Module):
    def __init__(self, img_size, embed_size, num_heads, mlp_dim, Conv, Norm):
        super().__init__()
        self.norm1 = nn.LayerNorm(embed_size)
        self.msa = _MSA(embed_size, num_heads)

        self.norm2 = nn.LayerNorm(embed_size)
        self.mlp = _MLP(embed_size, num_heads, Norm)

    def forward(self, x):
        y = self.norm1(x)
        y = self.msa(y)
        x = x + y

        y = self.norm2(x)
        y = self.mlp(y)
        x = x + y

        return x


class _Transformer(nn.Sequential):
    def __init__(
        self,
        img_size,
        patch_size,
        in_feats,
        embed_size,
        num_heads,
        mlp_dim,
        num_layers,
        out_layers,
        Conv,
        Norm,
    ):
        super().__init__()
        self.out_layers = out_layers
        self.num_layers = num_layers
        self.embed = _Embedding(img_size, patch_size, in_feats, embed_size, Conv)
        self.layers = ModuleList([])

        for _ in range(0, num_layers):
            layer = _TransformerLayer(
                img_size, embed_size, num_heads, mlp_dim, Conv, Norm
            )
            self.layers.append(layer)

    def forward(self, x):
        out = []
        x = self.embed(x)

        for i in range(0, self.num_layers):
            x = self.layers[i](x)
            if i in self.out_layers:
                out.append(x)

        return out


def checkImgSize(img_size, number=4):
    if all([x >= 2**number for x in img_size]):
        return number
    else:
        return int(np.min(np.floor(np.log2(img_size))))


class unetr(ModelBase):
    """
    This is the U-NetR architecture : https://arxiv.org/abs/2103.10504.
    The 'residualConnections' flag controls residual connections The Downsampling, Encoding, Decoding modules are defined in the seg_modules file.
    These smaller modules are basically defined by 2 parameters, the input channels (filters) and the output channels (filters),
    and some other hyperparameters, which remain constant all the modules. For more details on the smaller modules please have a look at the seg_modules file.
    """

    def __init__(
        self,
        parameters: dict,
    ):
        super(unetr, self).__init__(parameters)

        # initialize defaults if not found
        parameters["model"]["inner_patch_size"] = parameters["model"].get(
            "inner_patch_size", parameters["patch_size"][0]
        )
        parameters["model"]["num_heads"] = parameters["model"].get("num_heads", 12)
        parameters["model"]["embed_dim"] = parameters["model"].get("embed_dim", 768)

        self.patch_size = parameters["model"]["inner_patch_size"]

        assert np.ceil(np.log2(parameters["model"]["inner_patch_size"])) == np.floor(
            np.log2(parameters["model"]["inner_patch_size"])
        ), "The inner patch size must be a power of 2."

        self.depth = int(np.log2(self.patch_size))
        parameters["model"]["depth"] = self.depth

        _ = self.model_depth_check(parameters)

        if self.n_dimensions == 2:
            self.img_size = parameters["patch_size"][0:2]
        elif self.n_dimensions == 3:
            self.img_size = parameters["patch_size"]

        self.num_layers = 3 * self.depth  # number of transformer layers
        self.out_layers = np.arange(2, self.num_layers, 3)

        self.num_heads = parameters["model"]["num_heads"]
        self.embed_size = parameters["model"]["embed_dim"]

        self.patch_dim = [i // self.patch_size for i in self.img_size]

        assert (
            self.embed_size % self.num_heads == 0
        ), "The embedding dimension must be divisible by the number of self-attention heads"

        assert all(
            [i % self.patch_size == 0 for i in self.img_size]
        ), "The image size is not divisible by the patch size in at least 1 dimension. UNETR is not defined in this case."

        assert all(
            [self.patch_size <= i for i in self.img_size]
        ), "The inner patch size must be smaller than the input image."

        self.transformer = _Transformer(
            img_size=self.img_size,
            patch_size=self.patch_size,
            in_feats=self.n_channels,
            embed_size=self.embed_size,
            num_heads=self.num_heads,
            mlp_dim=2048,
            num_layers=self.num_layers,
            out_layers=self.out_layers,
            Conv=self.Conv,
            Norm=self.Norm,
        )

        self.upsampling = ModuleList([])
        self.convs = ModuleList([])

        for i in range(0, self.depth - 1):
            # add deconv blocks
            tempconvs = nn.Sequential()
            tempconvs.add_module(
                "conv0",
                _DeconvConvBlock(
                    self.embed_size,
                    32 * 2**self.depth,
                    self.Norm,
                    self.Conv,
                    self.ConvTranspose,
                ),
            )

            for j in range(self.depth - 2, i, -1):
                tempconvs.add_module(
                    "conv%d" % j,
                    _DeconvConvBlock(
                        128 * 2**j,
                        128 * 2 ** (j - 1),
                        self.Norm,
                        self.Conv,
                        self.ConvTranspose,
                    ),
                )

            self.convs.append(tempconvs)

            # add upsampling
            self.upsampling.append(
                _UpsampleBlock(
                    128 * 2 ** (i + 1), self.Norm, self.Conv, self.ConvTranspose
                )
            )

        # add upsampling for transformer output (no convs)
        self.upsampling.append(
            self.ConvTranspose(
                in_channels=self.embed_size,
                out_channels=32 * 2**self.depth,
                kernel_size=2,
                stride=2,
                padding=0,
                output_padding=0,
            )
        )

        self.input_conv = nn.Sequential()
        self.input_conv.add_module(
            "conv1", _ConvBlock(self.n_channels, 32, self.Norm, self.Conv)
        )
        self.input_conv.add_module("conv2", _ConvBlock(32, 64, self.Norm, self.Conv))

        self.output_conv = nn.Sequential()
        self.output_conv.add_module("conv1", _ConvBlock(128, 64, self.Norm, self.Conv))
        self.output_conv.add_module("conv2", _ConvBlock(64, 64, self.Norm, self.Conv))
        self.output_conv.add_module(
            "conv3",
            out_conv(
                64,
                self.n_classes,
                conv_kwargs={
                    "kernel_size": 1,
                    "stride": 1,
                    "padding": 0,
                    "bias": False,
                },
                norm=self.Norm,
                conv=self.Conv,
                final_convolution_layer=self.final_convolution_layer,
                sigmoid_input_multiplier=self.sigmoid_input_multiplier,
            ),
        )

    def forward(self, x):
        """
        Parameters
        ----------
        x : Tensor
            Should be a 5D Tensor as [batch_size, channels, x_dims, y_dims, z_dims].

        Returns
        -------
        x : Tensor
            Returns a 5D Output Tensor as [batch_size, n_classes, x_dims, y_dims, z_dims].

        """
        transformer_out = self.transformer(x)

        y = self.upsampling[-1](
            transformer_out[-1]
            .transpose(-1, -2)
            .view(-1, self.embed_size, *self.patch_dim)
        )  # z12

        for i in range(len(self.convs) - 1, -1, -1):
            zi = (
                transformer_out[i]
                .transpose(-1, -2)
                .view(-1, self.embed_size, *self.patch_dim)
            )
            zi = self.convs[i](zi)
            zicat = torch.cat([zi, y], dim=1)
            y = self.upsampling[i](zicat)

        x = self.input_conv(x)
        x = torch.cat([x, y], dim=1)
        x = self.output_conv(x)

        return x



from .wrap_torch import (
    sgd,
    asgd,
    adam,
    adamw,
    adamax,
    # sparseadam,
    rprop,
    adadelta,
    adagrad,
    rmsprop,
)

global_optimizer_dict = {
    "sgd": sgd,
    "asgd": asgd,
    "adam": adam,
    "adamw": adamw,
    "adamax": adamax,
    # "sparseadam": sparseadam,
    "rprop": rprop,
    "adadelta": adadelta,
    "adagrad": adagrad,
    "rmsprop": rmsprop,
}


def get_optimizer(params):
    """
    Function to get the optimizer definition.

    Args:
        params (dict): The parameters' dictionary.

    Returns:
        model (Optimizer): The optimizer definition.
    """
    return global_optimizer_dict[params["optimizer"]["type"]](params)
